#include "mpi.h"
#include <iostream>
#include <limits>
#include <vector>
#include <cstdlib>
#include <ctime>
#include <cstring>

struct Student {
    int id;
    char name[50];
    double grade;
};

const int num_first_names = 10;
const std::string first_names[] = {
        "John", "Alice", "Bob", "Eve", "Charlie", "Grace", "Oliver", "Sophia", "David", "Lily"
};
const int num_last_names = 10;
const std::string last_names[] = {
        "Smith", "Johnson", "Brown", "Davis", "Lee", "White", "Harris", "Clark", "Young", "Lewis"
};

struct Student find_best_student(struct Student students[], int num_students) {
    struct Student best_student;
    best_student.grade = 100;
    strcpy(best_student.name, "");

    for (int i = 0; i < num_students; i++) {
        if (students[i].grade < best_student.grade ||
            (students[i].grade == best_student.grade && strcmp(students[i].name, best_student.name) < 0)) {
            best_student = students[i];
        }
    }

    return best_student;
}

MPI_Datatype create_student_mpi_type() {

    // TODO implement:
    /*
    - Create custom MPI Datatype for sending data about Student
        * Datatype is derived from structure Student defined above
        * Create array of int representing length of each block of data
        * Create array of MPI_Datatype of proper size and set it values to proper MPI_Datatypes
        * Create array of MPI_Aint of proper size representing offset of each block
        * Get values of offset for each block using offsetof
        * Create structure and commit it
    */
}

MPI_Op create_student_min_op() {
    
    // TODO implement:
    /*
    - Create custom MPI operation student_min (MPI_Op) using MPI_Op_create for finding the best student
        * Function behaviour is the same as for find_best_student define above
        * You can use lambda function or create your own function above
        * Cast in and inout values to proper type
        * Iterate through array elements and use comparison define above to find the best student
        * NOTE : This function is cumulative
    */
}

int main(int argc, char** argv) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int num_students_per_process = 5;
    int total_students = num_students_per_process * size;

    Student *local_data = new Student[num_students_per_process];
    Student *all_students = nullptr;

    if (rank == 0) {
        all_students = new Student[total_students];
        std::cout << "Students: " << std::endl;
        srand(static_cast<unsigned>(time(0)));
        for (int i = 0; i < total_students; i++) {
            all_students[i].id = i;
            int first_name_index = rand() % num_first_names;
            int last_name_index = rand() % num_last_names;
            std::string full_name = last_names[last_name_index] + " " + first_names[first_name_index];
            strcpy(all_students[i].name, full_name.c_str());
            all_students[i].grade = rand() % 101;
            std::cout << "\tStudent - ID: "
                      << all_students[i].id << ", Name: "
                      << all_students[i].name << ", Grade: "
                      << all_students[i].grade << std::endl;
        }
    }

    MPI_Datatype student_mpi_type = create_student_mpi_type();

    MPI_Scatter(all_students, num_students_per_process, student_mpi_type, local_data, num_students_per_process,
                student_mpi_type, 0, MPI_COMM_WORLD);

    Student local_best_student;
    local_best_student.grade = std::numeric_limits<double>::max();
    local_best_student = find_best_student(local_data, num_students_per_process);

    Student global_best_student;
    global_best_student.grade = std::numeric_limits<double>::max();
    MPI_Op student_min = create_student_min_op();

    MPI_Reduce(&local_best_student, &global_best_student, 1, student_mpi_type, student_min, 0, MPI_COMM_WORLD);

    MPI_Op_free(&student_min);
    MPI_Type_free(&student_mpi_type);

    if (rank == 0) {
        std::cout << "\nBest student - ID: "
                  << global_best_student.id << ", Name: "
                  << global_best_student.name << ", Grade: "
                  << global_best_student.grade << "\n" << std::endl;
        delete[] all_students;
    }

    delete[] local_data;

    MPI_Finalize();
    return 0;
}