#include "mpi.h"
#include <stdio.h>

// Define a custom structure for a 2D point
struct Point2D {
    double x;
    double y;
};

int main(int argc, char **argv) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    // Create a custom MPI datatype for the Point2D structure
    MPI_Datatype point2d_type;
    MPI_Type_contiguous(2, MPI_DOUBLE, &point2d_type);
    MPI_Type_commit(&point2d_type);

    if (size < 2) {
        printf("This example requires at least two processes.\n");
        MPI_Finalize();
        return 1;
    }

    if (rank == 0) {
        struct Point2D point = {1.0, 2.0};

        // Process 0 sends the custom structure to process 1
        MPI_Send(&point, 1, point2d_type, 1, 0, MPI_COMM_WORLD);
        printf("Process 0 sent Point2D: (%.1f, %.1f) to Process 1.\n", point.x, point.y);
    } else if (rank == 1) {
        struct Point2D received_point;

        // Process 1 receives the custom structure from process 0
        MPI_Recv(&received_point, 1, point2d_type, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        printf("Process 1 received Point2D: (%.1f, %.1f) from Process 0.\n", received_point.x, received_point.y);
    }

    MPI_Type_free(&point2d_type);
    MPI_Finalize();

    return 0;
}
